#!/bin/bash    

if [ 1 -eq 0 ]; then
  #
  # useful commands
  #

  ./solver.py -v 20 -r 1 -R 4 -X 1 -U 1 -W 3 --Nxy 724
  at now + 2 hours ... CNTRL D

  cd ./results; while true; do date > atjob.touch && cp -vpru . ~/Dropbox/AccessAndShare/fenics/results/; sleep 3600; done
  
  chmod u+x bigrun.sh solver.py
  time ./bigrun.sh -J 19 | tee runmeout.txt ; cd ./results ; cp -vpru . ~/Dropbox/AccessAndShare/fenics/results/ ; cd .. ; date

  diff ./solver.py ~/Dropbox/AccessAndShare/fenics/solver.py ;
  diff ./bigrun.sh ~/Dropbox/AccessAndShare/fenics/bigrun.sh ;
  cp ./bigrun.sh ~/Dropbox/AccessAndShare/fenics/bigrun.sh ; 
  cp ./solver.py ~/Dropbox/AccessAndShare/fenics/solver.py ;
fi

# set up the defaults
MAXJ=4
VERBOSE=0

# parse the command line
while [[ $# -gt 0 ]]
do
  OPT="$1"
  shift
  
  case $OPT in
  
    -J|--maxj)
      MAXJ="$1"
      shift # past argument
    ;;
    
    -v|--verbose)
      VERBOSE=1
    ;;
    
    -p|--PDF)
      enscript --color=1 --margins 10c::: -C -f Courier8 -E -M A4 -p ./solver.ps solver.py
      ps2pdf ./solver.ps; rm solver.ps
      enscript --color=1 --margins 10c::: -C -f Courier8 -E -M A4 -p ./bigrun.ps bigrun.sh
      ps2pdf ./bigrun.ps; rm bigrun.ps
      exit 0
    ;;
    
    -b|--backup)
      DATESTAMP=`date +%Y_%m_%d_%H-%M-%S`
      echo "backing up to $DATESTAMP-solver.py and $DATESTAMP-bigrun.sh"
      cp solver.py ./backup/$DATESTAMP-solver.py
      cp bigrun.sh ./backup/$DATESTAMP-bigrun.sh
      exit 0
    ;;
    
    -h|--help)
      echo "Help text: e.g. time ./bigrun.sh -J 8 -v "
      echo "    -p (to create PDF's)"
      echo "    -b (to backup solver.py and bigrun.sh to backup directory)"
      echo "    -v (verbose)"
      echo "    -h"
      echo "    -J 19"
      echo "        This -J option corresponds, for example, to ..."
      for((j=2; j<=21; j+=2)); do
        echo "          -J $j => --Nxy `awk "BEGIN {print int(0.5+2^($j/2))}"` "
      done
      exit 0
    ;;
    
    *)
      echo "option $OPT unknown"
      exit 1 # for echo $?
    ;;
  esac
  
done


# report back if desired
if [ ${VERBOSE} -ne 0 ]; then
  echo MAXJ  = "${MAXJ}"
fi

# the original command
#rm -f output.txt; for((j=2; j<19; j++)); do ./solver.py -N `awk "BEGIN {print int(0.5+2^($j/2))}"` | tee -a output.txt; echo "grepping output"; cat output.txt | grep latex | tee errors.tex ; cat output.txt | grep matlab | tee errs.m ; done ;  echo ']' >> errs.m ; echo 'results = [' | cat - errs.m > ./errors.m ; rm errs.m; rm -rf ./results/X ; mkdir ./results/X ; cp ./solver.py ./results/X ; mv errors.* output.txt ./results/X

# define a bash function to do this with MAXJ and the results target directory as variables.
run_example ()
{
  # start fresh with a new results directory and store the run
  rm -fr ./results/$1
  mkdir ./results/$1
  # generate a new output file
  rm -f output.txt
  for((j=2; j<=$2; j++)); do
    # start a timer
    THEN=$SECONDS
    # intialise the runstats.m file
    echo 'runstats = [' > runstats.m ;
    # remove leading whitespace so as not to sometimes get a false field
    top -b -d 1.0 | sed 's/^ *//g' | grep "solver" | tr -s " " " "  | cut -d" " -f9,10,11 | sed -e 's/:/*60+/g' >> runstats.m &
    # build the argument list
    ARGLIST="-v 20 -i 100 -r 1 -R 4 -X $1 -U 1 -W 0 --Nxy `awk "BEGIN {print int(0.5+2^($j/2))}"` "
    # run the solver
    echo `date`
    echo "Running ./solver.py $ARGLIST " | tee -a output.txt
    python ./solver.py $ARGLIST | tee -a output.txt
    echo "grepping output";
    # extract latex data from the output
    cat output.txt | grep latex  | tee errors.tex ;
    # extract matlab data from the output
    cat output.txt | grep matlab | tee errs.m ;
    # tidy up runstats, start by nicely terminating top and sleep for a second
    kill `pidof top`
    sleep 1s
    echo '];' >> runstats.m
    # end the timer
    NOW=$SECONDS
    DURATION=$(( NOW - THEN ))
    echo "% Total runtime  (secs): $DURATION" | tee -a runstats.m
    echo "% Total runtime  (mins): $(bc <<<"scale=3; ($NOW-$THEN)/60.0")" | tee -a runstats.m
    echo "% Total runtime (hours): $(bc <<<"scale=3; ($NOW-$THEN)/3600.0")" | tee -a runstats.m
    mv runstats.m `echo runstats_J_eq_$j.m`
    mv runstats*.m ./results/$1
  done
  # clean up the matlab file with a variable name
  echo '];' >> errs.m
  echo 'results = [' | cat - errs.m > ./errors.m
  rm errs.m;
  # move these to the results directory
  cp ./solver.py ./results/$1
  mv errors.* output.txt ./results/$1
}

echo "% runnable " > runplotter.m
# this is the actual run. Loop through the examples and call the function above.
# note that jot works as well as seq - but might need installing
#for EXMPL in 12 ; do
for EXMPL in `seq 12 -1 1`; do
  run_example $EXMPL $MAXJ
  echo "clear all; close all; plotter($EXMPL)" >> runplotter.m 
done
echo "close all " >> runplotter.m
mv -f runplotter.m ./results/


