# SS #
#   !/usr/bin/python

"""
Based on demo_elastodynamics.py as distributed as part of DOLFIN.
solver10.py - first working version on 6 Feb 2017. Started on Feb 3rd.
Lots of trouble with getting the update assignment to work.
Working version with internal variables on 14 March 2017, pi day, but I forgot the
initial data in the right hand side. Final working version on 19 March 2017.

#turn off 3D acceleration in vbox settings to eliminate many open GL errors.

To Do:
1) Look at numerical energy error as a function of frequency
2) check which variables, trial functions etc are actually used
3) add -C 'a comment' to ./bigrun.sh command for description

Example:
date ; mpirun -np 2 ./solver.py -v 20 -r 1 -R 4 -X 6 -U 1 -W 5 --Nxy 128 -C '-W 5 with both tols 1e-10' ; date


./solver.py -v 20 -r 1 -R 4 -X 6 -U 1 -W 5 --Nxy 8 -C '-W 5 with both tols 1e-10'

   8 &  3.443e+01 &  2.333e-02 &  3.443e+01 &  2.648e+01 &  6.848e-01 \\ %  latex: dolfin v. 2016.2.0 

./runme_np.sh ssh 2> > ( tee output_np.txt )

The successful run for example 6 at 1024 had the #!# interpolates of IC's

"""
import math
from dolfin import *
# SS # 
import dolfin
from fenics import *
import os, sys, getopt, time
from mpi4py import MPI

#set_log_level(0) 
# SS # set_log_active(False)

# help function
def usage():
  set_log_active(True)
#  if MPI.COMM_WORLD.rank == 0:
  print("-h or --help")
  print("-v d or --garrulous for verbosity")
  print("-r d or --rdeg to specify FE polynomial degree")
  print("-R d or --Rdeg to specify 'exact solution' polynomial degree")
  print("-X d or --example to specify the example to run")
  print("-N n or --Nxy to specify Nxy")
  print("-n n or --Nt to specify Nt")
  print("-i n or --pvi to specify progvalinc - the increment for progress reports")
  print("-T T or --Tfinal T to specify final time, T")
  print("-C s or --Comment 'xxx yyy' to specify string s to append to output line")
  print("-W m or --Wmeth to specify solution method for W")
  print("-U m or --Umeth to specify solution method for U")
  print(" ")
  list_lu_solver_methods(); print(" ")
  list_krylov_solver_methods(); print(" ")
  list_krylov_solver_preconditioners(); print(" ")
  os.system('date +%Y_%m_%d_%H-%M-%S')
  print (time.strftime("%d/%m/%Y at %H:%M:%S"))
 
# parse the command line
try:
  opts, args = getopt.getopt(sys.argv[1:], "hr:R:N:n:i:T:C:X:v:U:W:",
                   [
                    "help",           # obvious
                    "garrulous=",     # level of verbosity
                    "rdeg=",          # FE polynomial degree
                    "Rdeg=",          # solution polynomial degree
                    "Nxy=",           # integer for N by N mesh
                    "Nt=",            # integer for Nt time steps
                    "pvi=",           # progvalinc - increment for progress output
                    "Tfinal=",        # final time, T
                    "Comment=",       # string to append to output
                    "Wmeth=",         # solution method for W
                    "Umeth=",         # solution method for U
                    "example="        # integer for example number
                    ])

except getopt.GetoptError as err:
  # print help information and exit:
  print(err) # will print something like "option -a not recognized"
  usage()
  sys.exit(2)

# set defaults
beingloud  = 0        # level of verbosity
Nxy        = 4        # default value for the mesh density
Nt         = 4        # default value for the number of time steps
example    = 1        # run example 1 by default
rdeg       = 1        # order of CG elements
Rdeg       = rdeg+3   # order of polynomial for exact solution
Wmethod    = 0;       # solution method for W
Umethod    = 0;       # solution method for U
Comment    = "";      # string to append to output
progvalinc = 16;       # output progress every 1/4 of Nt

# basic solution parameters, the final time...
T = 6.0*2.0*math.pi;
# mass density, Young's modulus, Poisson ratio and Lame coefficients
RHO     = 1.0 # 1000
E       = 1.0 # 100000
nu      = 0.45
MU      = 1.0 # E / (2.0*(1.0 + nu))
LAMBDA  = 1.0 # E*nu / ((1.0 + nu)*(1.0 - 2.0*nu))
# Rayleigh Damping: cannot take these to be zero in the forms - integral without a domain!
gM = 1.0e-10; gE = 1.0e-10
# Maxwell-Zener terms
tau1 = 0.1; tau2 = 0.05; varphi1 = 3.5e-10; varphi2 = 1.5e-10

#if MPI.COMM_WORLD.rank == 0:
print('Command Line: using: ')
for o, a in opts:
  if o in ("-v","--garrulous"):
    beingloud = int(a)
    if beingloud > 19: print('loud level %d;' % beingloud),
  elif o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-r", "--rdeg"):
    rdeg = int(a)
    if beingloud > 19: print('rdeg = %d;' % rdeg),
  elif o in ("-R", "--Rdeg"):
    Rdeg = int(a)
    if beingloud > 19: print('Rdeg = %d;' % Rdeg),
  elif o in ("-N", "--Nxy"):
    Nxy = int(a)
    if beingloud > 19: print('Nxy = %d;' % Nxy),
  elif o in ("-n", "--Nt"):
    Nt = int(a)
    if beingloud > 19: print('Nt = %d;' % Nt),
  elif o in ("-i", "--pvi"):
    progvalinc = int(a)
    if beingloud > 19: print('progvalinc = %d;' % progvalinc),
  elif o in ("-T", "--Tfinal"):
    T = float(a)
    if beingloud > 19: print('T = %f;' % T),
  elif o in ("-C", "--Comment"):
    Comment = str(a)
    if beingloud > 19: print('Comment = %s;' % Comment),
  elif o in ("-W", "--Wmeth"):
    Wmethod = int(a)
    if beingloud > 19: print('method %d for the W solve;' % Wmethod),
  elif o in ("-U", "--Umeth"):
    Umethod = int(a)
    if beingloud > 19: print('method %d for the U solve;' % Umethod),
  elif o in ("-X", "--example"):
    example = int(a)
    if beingloud > 19: print('example %d;' % example),
  else:
    assert False, "unhandled option"

if MPI.COMM_WORLD.rank == 0:
  print(""); print('Command line parsing complete...')

# uncomment and trial run times for these form compiler options
parameters["form_compiler"]["cpp_optimize"] = True
parameters["form_compiler"]["optimize"] = True

# define FE, mesh and time stepping parameters
BU = 0.0
tn = 0.0
#Nt = 4
k = T/Nt

# first modify default material property data
if   example == 0:
  gM = 20.0e-15; gE = 10.0e-15; tau1 = 0.1; tau2 = 0.05; varphi1 = 3.5e-15; varphi2 = 1.5e-15
elif example == 1 or example == 2 or example == 3 or example == 4:
  gM = 20.0e-15; gE = 10.0e-15; varphi1 = 3.5e-15; varphi2 = 1.5e-15
elif example == 5 or example == 6 or example == 7 or example == 8:
  gM = 20.0e-1;  gE = 10.0e-1;  varphi1 = 3.5e-15; varphi2 = 1.5e-15
elif example == 9 or example == 10 or example == 11 or example == 12:
  gM = 20.0e-1;  gE = 10.0e-1;  varphi1 = 3.5e-1;  varphi2 = 1.5e-1; tau1 = 0.1; tau2 = 0.05
else:
  print("Error!! Exiting ..."); exit(0)

# next set up the run time parameters
hexpo = -1.0;
if   example == 0:
  BU=0.0; hexpo -1.0
elif example == 1 or example == 5 or example == 9:
  BU=0.0; hexpo -1.0
elif example == 2 or example == 6 or example == 10:
  BU=1.0; hexpo = rdeg/1.5
elif example == 3 or example == 7 or example == 11:
  BU=1.0; hexpo = rdeg/3.0
elif example == 4 or example == 8 or example == 12:
  BU=1.0; hexpo = rdeg/6.0
else:
  print("Error!! Exiting ..."); exit(0)

if hexpo < 0:
  k = T/Nt # Nt = 4;
else:
  h=1.0/Nxy; k = math.pow(h,hexpo); Nt = max(1,int(T/k)); k=T/Nt

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Solving for example ", example)
    print("Solving with Nxy, k, T, T/k, Nt = ", Nxy, k, T, T/k, Nt)
    print("gM, gE, varphi1, varphi2, tau1, tau2 = ", gM, gE, varphi1, varphi2, tau1, tau2)
    print("BU, Nt, k = ", BU, Nt, k)
    # SS #print("dolfin version: ", dolfin.dolfin_version())

# finalise derived quantities
varphi0 = 1.0 - varphi1 - varphi2
beta1   = math.sqrt(varphi1*tau1)
beta2   = math.sqrt(varphi2*tau2)
# create some useful shorthands
zdenom1 = 6*tau1*tau1 + 4*k*tau1 + k*k
zdenom2 = 6*tau2*tau2 + 4*k*tau2 + k*k

# generate a mesh on the unit square
mesh = UnitSquareMesh(Nxy, Nxy)
# left edge
def left(x, on_boundary):
    return x[0] < 0.001 and on_boundary
# right edge
def right(x, on_boundary):
    return x[0] > 0.999 and on_boundary
# bottom edge
def bottom(x, on_boundary):
    return x[1] < 0.001 and on_boundary
# top edge
def top(x, on_boundary):
    return x[1] > 0.999 and on_boundary

# finite element space for displacements and velocities
V = VectorFunctionSpace(mesh, "CG", degree=rdeg) # dim of mesh is implied
# higher order FE space for 'exact' solution
U = VectorFunctionSpace(mesh, "CG", degree=Rdeg)

# spaces for initial data projections
ws, us = TrialFunction(V), TrialFunction(V)
vs = TestFunction(V)

# product spaces and functions for the time-slab solve

# first line needed on the later FEniCS versions; second on the earlier
V1 = VectorElement("CG", mesh.ufl_cell(), rdeg); VV = FunctionSpace(mesh, V1*V1)
#VV = V * V

w, u, z1, z2  = TrialFunction(VV), TrialFunction(VV), TrialFunction(VV), TrialFunction(VV)
v = TestFunction(VV)
# access components for use in the forms
u1, u2 = split(u); w1, w2 = split(w); v1, v2 = split(v)
z11, z12 = split(z1); z21, z22 = split(z2)
# current values of pairs of w and u vectors
wc, uc, z1c, z2c = Function(VV), Function(VV), Function(VV), Function(VV)
# previous time slab values for RHS forcing terms
wp1, wp2, up1, up2 = Function(V), Function(V), Function(V), Function(V)
zp11, zp12, zp21, zp22 = Function(V), Function(V), Function(V), Function(V)
# pretend, high order, exact functions
uexact, wexact = Function(U), Function(U)

# for the iterative solvers
wA, wAp, wB, wBp = Function(V), Function(V), Function(V), Function(V)

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Functions and spaces now defined")

# useful constant
zero = Constant((0.0, 0.0))
# define exact solutions
Uex   = Expression( ("16.0*(x[0]*x[0]-x[0])*(x[1]*x[1]-x[1])*(  t+B*cos(t))",                 \
                     "16.0*(x[0]-x[0]*x[0])*(x[1]-x[1]*x[1])*(  t+B*cos(t))"),                \
                      degree=5, t=0, B=BU)
Wex   = Expression( ("16.0*(x[0]*x[0]-x[0])*(x[1]*x[1]-x[1])*(1.0-B*sin(t))",                 \
                     "16.0*(x[0]-x[0]*x[0])*(x[1]-x[1]*x[1])*(1.0-B*sin(t))"),                \
                      degree=5, t=0, B=BU)

# f with time dependence...
f1xy = Expression(("RHO*6.0*( B*sin(tn-k)-B*sin(tn)+gamM*(k+B*cos(tn)-B*cos(tn-k) ) )         \
                           *16.0*(x[0]*x[0]-x[0])*(x[1]*x[1]-x[1])                            \
                    +6.0*(k*(tn-0.5*k)+B*(sin(tn)-sin(tn-k))+gamE*(k+B*cos(tn)-B*cos(tn-k))   \
                    - varphi1*(tau1*tau1*( exp(-(tn-k)/tau1) - exp(-tn/tau1) )                \
                               + k*( (tn-k/2.0)-tau1 )                                        \
                               + B/(tau1*tau1+1)                                              \
                                 *( tau1*( exp(-tn/tau1) - exp(-(tn-k)/tau1) )                \
                                   +tau1*( cos(tn-k)-cos(tn) )-sin(tn-k)+sin(tn)              \
                                  )                                                           \
                              )                                                               \
                    - varphi2*(tau2*tau2*( exp(-(tn-k)/tau2) - exp(-tn/tau2) )                \
                               + k*( (tn-k/2.0)-tau2 )                                        \
                               + B/(tau2*tau2+1)                                              \
                                 *( tau2*( exp(-tn/tau2) - exp(-(tn-k)/tau2) )                \
                                   +tau2*( cos(tn-k)-cos(tn) )-sin(tn-k)+sin(tn)              \
                                  )                                                           \
                              )                                                               \
                     )                                                                        \
                   *(-16.0*2.0*(LAMBDA+2*MU)*(x[1]*x[1]-x[1])                                 \
                     -16.0*(LAMBDA+MU)*(1.0-2.0*x[0])*(1.0-2.0*x[1])                          \
                     -16.0*2*MU*(x[0]*x[0]-x[0]))                                             \
                   + 6.0*B*(                                                                  \
                          tau1*varphi1*(exp(-tn/tau1)-exp(-(tn-k)/tau1))                      \
                        + tau2*varphi2*(exp(-tn/tau2)-exp(-(tn-k)/tau2))                      \
                       )                                                                      \
                   *(-16.0*2.0*(LAMBDA+2*MU)*(x[1]*x[1]-x[1])                                 \
                     -16.0*(LAMBDA+MU)*(1.0-2.0*x[0])*(1.0-2.0*x[1])                          \
                     -16.0*2*MU*(x[0]*x[0]-x[0])                                              \
                     )",                                                                      \
                                                                                              \
                   "RHO*6.0*( B*sin(tn-k)-B*sin(tn)+gamM*(k+B*cos(tn)-B*cos(tn-k) ) )         \
                           *16.0*(x[0]-x[0]*x[0])*(x[1]-x[1]*x[1])                            \
                    +6.0*(k*(tn-0.5*k)+B*(sin(tn)-sin(tn-k))+gamE*(k+B*cos(tn)-B*cos(tn-k))   \
                    - varphi1*(tau1*tau1*( exp(-(tn-k)/tau1) - exp(-tn/tau1) )                \
                               + k*( (tn-k/2.0)-tau1 )                                        \
                               + B/(tau1*tau1+1)                                              \
                                 *( tau1*( exp(-tn/tau1) - exp(-(tn-k)/tau1) )                \
                                   +tau1*( cos(tn-k)-cos(tn) )-sin(tn-k)+sin(tn)              \
                                  )                                                           \
                              )                                                               \
                    - varphi2*(tau2*tau2*( exp(-(tn-k)/tau2) - exp(-tn/tau2) )                \
                               + k*( (tn-k/2.0)-tau2 )                                        \
                               + B/(tau2*tau2+1)                                              \
                                 *( tau2*( exp(-tn/tau2) - exp(-(tn-k)/tau2) )                \
                                   +tau2*( cos(tn-k)-cos(tn) )-sin(tn-k)+sin(tn)              \
                                  )                                                           \
                              )                                                               \
                     )                                                                        \
                    *( 16.0*2.0*(LAMBDA+2*MU)*(x[0]-x[0]*x[0])                                \
                      -16.0*(LAMBDA+MU)*(2.0*x[0]-1.0)*(2.0*x[1]-1.0)                         \
                      +16.0*2*MU*(x[1]-x[1]*x[1]) )                                           \
                   + 6.0*B*(                                                                      \
                          tau1*varphi1*(exp(-tn/tau1)-exp(-(tn-k)/tau1))                      \
                        + tau2*varphi2*(exp(-tn/tau2)-exp(-(tn-k)/tau2))                      \
                       )                                                                      \
                   *(16.0*2.0*(LAMBDA+2*MU)*(x[0]-x[0]*x[0])                                  \
                      -16.0*(LAMBDA+MU)*(2.0*x[0]-1.0)*(2.0*x[1]-1.0)                         \
                      +16.0*2*MU*(x[1]-x[1]*x[1])                                             \
                      )"),                                                                    \
           degree = 5, t=0.0, LAMBDA=LAMBDA, MU=MU, RHO=RHO,                                  \
           k=0.0, tn=0.0, B=BU, gamM=gM, gamE=gE,                                             \
           varphi1=varphi1, tau1=tau1, varphi2=varphi2, tau2=tau2)

f2xy = Expression(("RHO*( B*6.0/k*(k*sin(tn-k)+cos(tn)-cos(tn-k))                             \
                           +   3.0*gamM*k+6.0*gamM*B/k*(sin(tn)-sin(tn-k)-k*cos(tn-k)) )      \
                   *16.0*(x[0]*x[0]-x[0])*(x[1]*x[1]-x[1])                                    \
                     +(k*(3.0*tn-2.0*k) + B*6.0/k*(cos(tn-k)-cos(tn)-k*sin(tn-k))             \
                         +3.0*gamE*k + 6.0*gamE*B/k*(sin(tn)-sin(tn-k)-k*cos(tn-k))           \
                    - 6*varphi1*( (                                                           \
                                 6*exp(-(tn-k)/tau1)*tau1*tau1*(tau1-k)                       \
                                 -2*(tn-k)*(tn-k)*(tn-k)+3*(tn-k)*(tn-k)*tn                   \
                                 +3*tau1*(tn-k)*((tn-k)-2*tn)                                 \
                                 -6*tau1*tau1*tau1*exp(-tn/tau1)-tn*tn*(tn-3*tau1)            \
                                )/(-6.0*k)                                                    \
                                +B/k/(tau1*tau1+1)                                            \
                                *(tau1*tau1*( exp(-(tn-k)/tau1)-exp(-tn/tau1) )               \
                                 +tau1*(sin(tn-k)-sin(tn))+cos(tn-k)-cos(tn)                  \
                                 )                                                            \
                                +B/(tau1*tau1+1)                                              \
                                *(tau1*cos(tn-k)-sin(tn-k)-tau1*exp(-(tn-k)/tau1)             \
                                 )                                                            \
                              )                                                               \
                    - 6*varphi2*( (                                                           \
                                 6*exp(-(tn-k)/tau2)*tau2*tau2*(tau2-k)                       \
                                 -2*(tn-k)*(tn-k)*(tn-k)+3*(tn-k)*(tn-k)*tn                   \
                                 +3*tau2*(tn-k)*((tn-k)-2*tn)                                 \
                                 -6*tau2*tau2*tau2*exp(-tn/tau2)-tn*tn*(tn-3*tau2)            \
                                )/(-6.0*k)                                                    \
                                +B/k/(tau2*tau2+1)                                            \
                                *(tau2*tau2*( exp(-(tn-k)/tau2)-exp(-tn/tau2) )               \
                                 +tau2*(sin(tn-k)-sin(tn))+cos(tn-k)-cos(tn)                  \
                                 )                                                            \
                                +B/(tau2*tau2+1)                                              \
                                *(tau2*cos(tn-k)-sin(tn-k)-tau2*exp(-(tn-k)/tau2)             \
                                 )                                                            \
                              )                                                               \
                     )                                                                        \
                     *(-16.0*2.0*(LAMBDA+2*MU)*(x[1]*x[1]-x[1])                               \
                       -16.0*(LAMBDA+MU)*(1.0-2.0*x[0])*(1.0-2.0*x[1])                        \
                       -16.0*2*MU*(x[0]*x[0]-x[0]) )                                          \
                     +6.0*B*(                                                                 \
                          tau1*varphi1*exp(-(tn-k)/tau1)*(tau1/k*(1.0-exp(-k/tau1))-1)        \
                         +tau2*varphi2*exp(-(tn-k)/tau2)*(tau2/k*(1.0-exp(-k/tau2))-1)        \
                        )                                                                     \
                     *(-16.0*2.0*(LAMBDA+2*MU)*(x[1]*x[1]-x[1])                               \
                       -16.0*(LAMBDA+MU)*(1.0-2.0*x[0])*(1.0-2.0*x[1])                        \
                       -16.0*2*MU*(x[0]*x[0]-x[0]) )",                                        \
                                                                                              \
                   "RHO*( B*6.0/k*(k*sin(tn-k)+cos(tn)-cos(tn-k))                             \
                           +   3.0*gamM*k+6.0*gamM*B/k*(sin(tn)-sin(tn-k)-k*cos(tn-k)) )      \
                   *16.0*(x[0]-x[0]*x[0])*(x[1]-x[1]*x[1])                                    \
                     +(k*(3.0*tn-2.0*k) + B*6.0/k*(cos(tn-k)-cos(tn)-k*sin(tn-k))             \
                         +3.0*gamE*k + 6.0*gamE*B/k*(sin(tn)-sin(tn-k)-k*cos(tn-k))           \
                    - 6*varphi1*( (                                                           \
                                 6*exp(-(tn-k)/tau1)*tau1*tau1*(tau1-k)                       \
                                 -2*(tn-k)*(tn-k)*(tn-k)+3*(tn-k)*(tn-k)*tn                   \
                                 +3*tau1*(tn-k)*((tn-k)-2*tn)                                 \
                                 -6*tau1*tau1*tau1*exp(-tn/tau1)-tn*tn*(tn-3*tau1)            \
                                )/(-6.0*k)                                                    \
                                +B/k/(tau1*tau1+1)                                            \
                                *(tau1*tau1*( exp(-(tn-k)/tau1)-exp(-tn/tau1) )               \
                                 +tau1*(sin(tn-k)-sin(tn))+cos(tn-k)-cos(tn)                  \
                                 )                                                            \
                                +B/(tau1*tau1+1)                                              \
                                *(tau1*cos(tn-k)-sin(tn-k)-tau1*exp(-(tn-k)/tau1)             \
                                 )                                                            \
                              )                                                               \
                    - 6*varphi2*( (                                                           \
                                 6*exp(-(tn-k)/tau2)*tau2*tau2*(tau2-k)                       \
                                 -2*(tn-k)*(tn-k)*(tn-k)+3*(tn-k)*(tn-k)*tn                   \
                                 +3*tau2*(tn-k)*((tn-k)-2*tn)                                 \
                                 -6*tau2*tau2*tau2*exp(-tn/tau2)-tn*tn*(tn-3*tau2)            \
                                )/(-6.0*k)                                                    \
                                +B/k/(tau2*tau2+1)                                            \
                                *(tau2*tau2*( exp(-(tn-k)/tau2)-exp(-tn/tau2) )               \
                                 +tau2*(sin(tn-k)-sin(tn))+cos(tn-k)-cos(tn)                  \
                                 )                                                            \
                                +B/(tau2*tau2+1)                                              \
                                *(tau2*cos(tn-k)-sin(tn-k)-tau2*exp(-(tn-k)/tau2)             \
                                 )                                                            \
                              )                                                               \
                     )                                                                        \
                      *( 16.0*2.0*(LAMBDA+2*MU)*(x[0]-x[0]*x[0])                              \
                        -16.0*(LAMBDA+MU)*(2.0*x[0]-1.0)*(2.0*x[1]-1.0)                       \
                        +16.0*2*MU*(x[1]-x[1]*x[1]) )                                         \
                     +6.0*B*(                                                                 \
                          tau1*varphi1*exp(-(tn-k)/tau1)*(tau1/k*(1.0-exp(-k/tau1))-1)        \
                         +tau2*varphi2*exp(-(tn-k)/tau2)*(tau2/k*(1.0-exp(-k/tau2))-1)        \
                        )                                                                     \
                     *( 16.0*2.0*(LAMBDA+2*MU)*(x[0]-x[0]*x[0])                               \
                        -16.0*(LAMBDA+MU)*(2.0*x[0]-1.0)*(2.0*x[1]-1.0)                       \
                        +16.0*2*MU*(x[1]-x[1]*x[1]) )"),                                      \
           degree = 5, t=0.0, LAMBDA=LAMBDA, MU=MU, RHO=RHO,                                  \
           k=0.0, tn=0.0, B=BU, gamM=gM, gamE=gE,                                             \
           varphi1=varphi1, tau1=tau1, varphi2=varphi2, tau2=tau2)

# linear strain tensor
def epsilon(v):
  return 0.5*( grad(v) + grad(v).T )

# isotropic stress tensor
def sigma(v):
  return 2.0*MU*epsilon(v) + LAMBDA*tr(epsilon(v))*Identity(v.geometric_dimension())

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Data now defined")

# initialise
Uex.t = 0.0
Wex.t = 0.0
# Define boundary conditions for the initial data
bcLu  = DirichletBC(V, zero, left)
bcRu  = DirichletBC(V, zero, right)
bcTu  = DirichletBC(V, zero, top)
bcBu  = DirichletBC(V, zero, bottom)
# define the ensemble bc's on u for t=0
bc0u = [bcLu, bcRu, bcTu, bcBu]
# forms and solve for initial data
aw0 = RHO*inner(ws,vs)*dx; Lw0 = RHO*inner(Wex,vs)*dx; solve(aw0 == Lw0, wp1)
#au0 = inner(us,vs)*dx; Lu0 = inner(Uex,vs)*dx; solve(au0 == Lu0, up1)
uexact = interpolate(Uex, U)
au0 = inner(sigma(us),epsilon(vs))*dx
Lu0 = inner(sigma(uexact),epsilon(vs))*dx
solve(au0 == Lu0, up1, bc0u)

# quick and dirty: replace solve with #!# above
#!# ws = interpolate(Wex, V)
#!# us = interpolate(Uex, V)

# WHAT SHOULD BE DONE HERE?
#### initialise the hanging functions
###assign(wp2, Constant(0.0))
###assign(up2, Constant(0.0))

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Initial data now available")

# Define boundary conditions for the time stepping
bcL0  = DirichletBC(VV.sub(0), zero, left)
bcR0  = DirichletBC(VV.sub(0), zero, right)
bcT0  = DirichletBC(VV.sub(0), zero, top)
bcB0  = DirichletBC(VV.sub(0), zero, bottom)
# and for W1 - might be difficult with current definitions.
bcL1  = DirichletBC(VV.sub(1), zero, left)
bcR1  = DirichletBC(VV.sub(1), zero, right)
bcT1  = DirichletBC(VV.sub(1), zero, top)
bcB1  = DirichletBC(VV.sub(1), zero, bottom)
# define the ensemble bc's on both temporal DOF's for u and w
bcs = [bcL0, bcL1, bcR0, bcR1, bcT0, bcT1, bcB0, bcB1]

# forms for time evolution of w (tidier that original)
A11 = 3.0*k*k*varphi0     + 6.0*k*gE + k*k*beta1*beta1/zdenom1*6.0*(6*tau1+k) - k*k*beta1*beta1/zdenom1*3.0*6*tau1 + k*k*beta2*beta2/zdenom2*6.0*(6*tau2+k) - k*k*beta2*beta2/zdenom2*3.0*6*tau2
A12 = 2.0*k*k*varphi0     + 3.0*k*gE + k*k*beta1*beta1/zdenom1*3.0*(k-2*tau1) + k*k*beta2*beta2/zdenom2*6.0*3*tau2 + k*k*beta1*beta1/zdenom1*6.0*3*tau1     + k*k*beta2*beta2/zdenom2*3.0*(k-2*tau2) 
A21 =     k*k*varphi0     + 3.0*k*gE + k*k*beta1*beta1/zdenom1*3.0*(6*tau1+k) - k*k*beta1*beta1/zdenom1*2.0*6*tau1 + k*k*beta2*beta2/zdenom2*3.0*(6*tau2+k) - k*k*beta2*beta2/zdenom2*2.0*6*tau2     
A22 = 5.0*k*k/6.0*varphi0 + 2.0*k*gE + k*k*beta1*beta1/zdenom1*3.0*3*tau1 + k*k*beta1*beta1/zdenom1*2.0*(k-2*tau1) + k*k*beta2*beta2/zdenom2*3.0*3*tau2 + k*k*beta2*beta2/zdenom2*2.0*(k-2*tau2) 

B11 = 6.0*RHO + 6.0*k*gM*RHO
B12 =           3.0*k*gM*RHO
B21 = 6.0*RHO + 3.0*k*gM*RHO
B22 = 3.0*RHO + 2.0*k*gM*RHO

aw =                                                                                  \
       B11*inner(w1, v1)*dx                  + B12*inner(w2, v1)*dx                   \
     + B21*inner(w1, v2)*dx                  + B22*inner(w2, v2)*dx                   \
                                                                                      \
     + A11 *inner(sigma(w1), epsilon(v1))*dx + A12 *inner(sigma(w2), epsilon(v1))*dx  \
     + A21 *inner(sigma(w1), epsilon(v2))*dx + A22 *inner(sigma(w2), epsilon(v2))*dx

Lw = inner(f1xy, v1)*dx  +  inner(f2xy, v2)*dx                                   \
    + 6.0*RHO*inner(wp1,v1)*dx - 6.0*k*varphi0*inner(sigma(up1), epsilon(v1))*dx \
    + 6.0*RHO*inner(wp1,v2)*dx - 3.0*k*varphi0*inner(sigma(up1), epsilon(v2))*dx \
                                                                                 \
    - 6.0*k*beta1*tau1/zdenom1*(6*tau1+k)*inner(sigma(zp11), epsilon(v1))*dx     \
    - 6.0*k*beta1*tau1/zdenom1*(3*tau1+k)*inner(sigma(zp11), epsilon(v2))*dx     \
    - 6.0*k*beta2*tau2/zdenom2*(6*tau2+k)*inner(sigma(zp21), epsilon(v1))*dx     \
    - 6.0*k*beta2*tau2/zdenom2*(3*tau2+k)*inner(sigma(zp21), epsilon(v2))*dx

# forms for time evolution of u
au = varphi0*inner(sigma(u1),epsilon(v1))*dx                                     \
     + varphi0*inner(sigma(u2),epsilon(v2))*dx
Lu = varphi0*inner(sigma(k*(wp1+0.5*wp2)+up1),epsilon(v1))*dx                    \
     + varphi0*inner(sigma(-k*(3*wp1+wp2)/3),epsilon(v2))*dx
# be careful of the above. It needs the latest w but the previous u

# forms for the iterative methods - method 20 and beyond

# Define W1, W2 boundary conditions for the time stepping
bcL  = DirichletBC(V, zero, left)
bcR  = DirichletBC(V, zero, right)
bcT  = DirichletBC(V, zero, top)
bcB  = DirichletBC(V, zero, bottom)
bcw1m20 = [bcL, bcR, bcT, bcB]

# forms for the W1 equation
aw1m20 = B11*inner(ws, vs)*dx + A11*inner(sigma(ws), epsilon(vs))*dx
Lw1m20 = inner(f1xy, vs)*dx                                                      \
    + 6.0*RHO*inner(wp1,vs)*dx - 6.0*k*varphi0*inner(sigma(up1), epsilon(vs))*dx \
    - 6.0*k*beta1*tau1/zdenom1*(6*tau1+k)*inner(sigma(zp11), epsilon(vs))*dx     \
    - 6.0*k*beta2*tau2/zdenom2*(6*tau2+k)*inner(sigma(zp21), epsilon(vs))*dx     \
    - A12 *inner(sigma(wBp), epsilon(vs))*dx- B12*inner(wBp, vs)*dx

# forms for the W2 equation
aw2m20 = B22*inner(ws, vs)*dx + A22*inner(sigma(ws), epsilon(vs))*dx
Lw2m20 = inner(f2xy, vs)*dx                                                      \
    + 6.0*RHO*inner(wp1,vs)*dx - 3.0*k*varphi0*inner(sigma(up1), epsilon(vs))*dx \
    - 6.0*k*beta1*tau1/zdenom1*(3*tau1+k)*inner(sigma(zp11), epsilon(vs))*dx     \
    - 6.0*k*beta2*tau2/zdenom2*(3*tau2+k)*inner(sigma(zp21), epsilon(vs))*dx     \
    - B21*inner(wAp, vs)*dx - A21 *inner(sigma(wAp), epsilon(vs))*dx

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Beginning time-stepping")

n=1
progval = 0
tic = time.time()
solve_time = time.time();
while n <= Nt:
  tn = n*k; Uex.t = tn; Wex.t = tn; # f.t = tn
  f1xy.user_parameters["k"]  = k
  f1xy.user_parameters["tn"] = tn
  f2xy.user_parameters["k"]  = k
  f2xy.user_parameters["tn"] = tn

  # solve for w
  if   Wmethod == 0:
    # safe, but requires too much memory for dense meshes
    solve(aw==Lw, wc, bcs)
  elif Wmethod == 1:
    # this seems non-convergent for dense meshes... Tolerances?
    solve(aw==Lw, wc, bcs, solver_parameters={"linear_solver": "gmres"})
  elif Wmethod == 2:
    # this seems to reproduce the non-convergent gmres above
    problem = LinearVariationalProblem(aw,Lw,wc,bcs)
    solver = LinearVariationalSolver(problem)
    prm = solver.parameters
    prm['linear_solver'] = 'gmres'
    prm['preconditioner'] = 'ilu'
    solver.solve()
  elif Wmethod == 3:
    # so let's play with the tolerances
    problem = LinearVariationalProblem(aw,Lw,wc,bcs)
    solver = LinearVariationalSolver(problem)
    prm = solver.parameters
    prm['linear_solver'] = 'gmres'
    prm['preconditioner'] = 'ilu'
    #prm['preconditioner'] = 'hypre_amg'
    #prm['krylov_solver']['absolute_tolerance']  = 1e-5     # default 1e-9 ?
    #prm['krylov_solver']['relative_tolerance']  = 1e-5     # default 1e-7 ?
    #prm['krylov_solver']["maximum_iterations"]  = 10000   # default 1000?
    #prm['krylov_solver']["monitor_convergence"] = True   # or False
    solver.solve()
  elif Wmethod == 4:
    # or experiment with other methods
    problem = LinearVariationalProblem(aw,Lw,wc,bcs)
    solver = LinearVariationalSolver(problem)
    prm = solver.parameters
    prm['linear_solver'] = 'umfpack'  # memory hungry for dense meshes
    solver.solve()
  elif Wmethod == 5:
    # or experiment with other methods
    problem = LinearVariationalProblem(aw,Lw,wc,bcs)
    solver = LinearVariationalSolver(problem)
    prm = solver.parameters
    prm['linear_solver'] = 'bicgstab' # appears best for memory
    # new - testing for bad 724 result: successful so retain for the future
    prm['krylov_solver']['absolute_tolerance']  = 1e-10     # default 1e-9 ?
    prm['krylov_solver']['relative_tolerance']  = 1e-10     # default 1e-7 ?
    #prm['krylov_solver']["maximum_iterations"]  = 10000   # default 1000?
    #prm['krylov_solver']["monitor_convergence"] = True   # or False
    solver.solve()
  elif Wmethod == 20:
    # first iterative solver
    TOL=1e-5; diff = 2*TOL; maxits = 1500; countit = 0
    # initial estimate for w
    assign(wAp,wp1)
    assign(wBp,wp2)
    while ( diff > TOL ):
      countit += 1; diff = 0
      assert countit <= maxits, 'maxits exceeded in solver...'
#      if countit > maxits:
#        print("maxits exceeded in solver... Exiting... IS COMMENTED"); # exit(0)
      # solve for wB - because we don't know wp2 at the first time step
      temp_choice = 5
      if temp_choice == 1:
        solve(aw2m20==Lw2m20, wB, bcw1m20)
      elif temp_choice == 2:
        solve(aw2m20==Lw2m20, wB, bcw1m20, solver_parameters={"linear_solver": "cg","preconditioner": "ilu"})
      else:
        problemB = LinearVariationalProblem(aw2m20,Lw2m20,wB,bcw1m20)
        solverB = LinearVariationalSolver(problemB)
        if temp_choice == 3:
          solverB.parameters['linear_solver'] = 'mumps'
        elif temp_choice == 4:
          solverB.parameters['linear_solver'] = 'petsc'
        elif temp_choice == 5:
          solverB.parameters['linear_solver'] = 'cg'
          solverB.parameters['symmetric'] = True
          solverB.parameters['preconditioner'] = 'icc'
          solverB.parameters['krylov_solver']['absolute_tolerance']  = 1e-9     # default 1e-9 ?
        else:
          print("temp_choice invalid in iterative Wmethod")
        solverB.solve()
        
      diff = diff + errornorm(wB, wBp, 'L2')
      assign(wBp,wB)
      #
      # REMARK: if we ever find a 'fast enough' solver we need to replicate the above here
      #
      # solve for wA
      solve(aw1m20==Lw1m20, wA, bcw1m20)
      diff = diff + errornorm(wA, wAp, 'L2')
      assign(wAp,wA)

  else:
    print("Error! No valid W solver option...  Exiting ..."); exit(0)

  if Wmethod < 20:
    assign(wp1,wc.sub(0))
    assign(wp2,wc.sub(1))
  else:
    assign(wp1,wA)
    assign(wp2,wB)
  
  
  if   Umethod == 0:
    # solve for u the long way
    solve(au==Lu, uc, bcs)
    assign(up1,uc.sub(0))
    assign(up2,uc.sub(1))
  elif Umethod == 1:
    # or this faster way
    up1.assign(k*(wp1+0.5*wp2)+up1)
    up2.assign(-k*(3*wp1+wp2)/3)
  else:
    print("Error! No valid U solver option...  Exiting ..."); exit(0)

  # update internal variables
  zp11.assign(2*tau1/zdenom1*(3*tau1-k)*zp11 \
               + k*beta1/zdenom1*((6*tau1+k)*wp1 + 3*tau1*wp2))
  zp12.assign(2*tau1/zdenom1*   3*k    *zp11 \
               - k*beta1/zdenom1*(    6*tau1  *wp1 + (2*tau1 -   k)*wp2))
  zp21.assign(2*tau2/zdenom2*(3*tau2-k)*zp21 \
               + k*beta2/zdenom2*((6*tau2+k)*wp1 + 3*tau2*wp2))
  zp22.assign(2*tau2/zdenom2*   3*k    *zp21 \
               - k*beta2/zdenom2*(    6*tau2  *wp1 + (2*tau2 -   k)*wp2))
  
  if 0:
    u0x, u0y = up1.split(); plot(u0x, interactive=False, key='u0x')
    w0x, w0y = wp1.split(); plot(w0x, interactive=False, key='w0x')
    plot(up1, mode="displacement", interactive=False, key='disp')

  if MPI.COMM_WORLD.rank == 0:
    if beingloud > 9 and n>=1.0*progval*Nt/progvalinc:
      toc = time.time()
      lap = toc-tic
      print("| t=%.4f  n/Nt=%4d/%-4d  | %3d%% | elapsed: %.2f (s), %.2f (m), %.2f (h), %.3f (d), %s, %s" 
             % (tn, n, Nt, 100*n/Nt, lap, lap/60.0, lap/3600.0, lap/86400.0, time.strftime("%d/%m/%Y at %H:%M:%S"), Comment ))
      sys.stdout.flush() # if not python3
      progval = progval+1
      tic = time.time()
  n = n+1;

# at the end of the time loop  
solve_time = time.time() - solve_time

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Time loop has now finished - at the endgame")

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Solve Time = ", solve_time)

if MPI.COMM_WORLD.rank == 0:
  if 0:
    if beingloud > 19: print("Plotting solutions at final time")
    # Plot solution
    #plot(mesh, interactive=False)
    u0x, u0y = up1.split(); plot(u0x, interactive=False, key='u0x')
    w0x, w0y = wp1.split(); plot(w0x, interactive=False,  key='w0x')

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19: print("Computing errors at final time")
  
if 1:
  # compute errors
  Uex.t = T
  Wex.t = T
  # L2, kinetic energy error
  u_err_L2 = sqrt(RHO)*errornorm(Uex, up1, 'L2') #sqrt( RHO*assemble((Uex-up1)**2*dx(mesh)) )
  w_err_L2 = sqrt(RHO)*errornorm(Wex, wp1, 'L2') #sqrt( RHO*assemble((Wex-wp1)**2*dx(mesh)) )
  #u_err_L2 = sqrt( RHO*assemble((Uex-up1)**2*dx(mesh), \
                                #form_compiler_parameters={'quadrature_degree': 8,'quadrature_scheme': 'vertex'}) )
  #w_err_L2 = sqrt( RHO*assemble((Wex-wp1)**2*dx(mesh), \
                                #form_compiler_parameters={'quadrature_degree': 8,'quadrature_scheme': 'vertex'}) )

  # mimic the exact solution in a higher-order space
  uexact = interpolate(Uex, U)
  wexact = interpolate(Wex, U)
  # H1, elastic-energy norm #help(errornorm) says this might be unstable
  u_err_En = 0.0
  u_err_En = sqrt( varphi0*assemble( inner(sigma(up1-uexact),epsilon(up1-uexact) )*dx(mesh)) )
  w_err_En = 0.0
  w_err_En = sqrt( varphi0*assemble( inner(sigma(wp1-wexact),epsilon(wp1-wexact) )*dx(mesh)) )
  # H1, Sobolev norm
  u_err_H1 = errornorm(Uex, up1, 'H1')
  w_err_H1 = errornorm(Wex, wp1, 'H1')
  # problem-defined energy norm
  energy_norm = sqrt( w_err_L2*w_err_L2 + u_err_En*u_err_En )
  # http://www.python-course.eu/python3_formatted_output.php

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print(" Nxy &     En     &     L2w    &    SEu     &    H1u     &    H1w     \\\\\\hline")
    print("{0:4d} & {1:10.3e} & {2:10.3e} & {3:10.3e} & {4:10.3e} & {5:10.3e} \\\\ %  latex: dolfin v. {6:s}; {7:s}, {8:s} "\
# SS #        .format(Nxy, energy_norm, w_err_L2, u_err_En, u_err_H1, w_err_H1, dolfin.dolfin_version(),
        .format(Nxy, energy_norm, w_err_L2, u_err_En, u_err_H1, w_err_H1, dolfin.__version__,
                time.strftime("%d/%m/%Y at %H:%M:%S"), Comment ) )
    print("{0:4d} , {1:10.3e} , {2:10.3e} , {3:10.3e} , {4:10.3e} , {5:10.3e} ;  % matlab: dolfin v. {6:s}; {7:s}, {8:s} "\
# SS #        .format(Nxy, energy_norm, w_err_L2, u_err_En, u_err_H1, w_err_H1, dolfin.dolfin_version(),
        .format(Nxy, energy_norm, w_err_L2, u_err_En, u_err_H1, w_err_H1, dolfin.__version__,
                time.strftime("%d/%m/%Y at %H:%M:%S"), Comment ) )
else:
  pass

if MPI.COMM_WORLD.rank == 0:
  if beingloud > 19:
    print("Ended successfully")

exit(0)
